/*
 * kinetic-cpp-client
 * Copyright (C) 2014 Seagate Technology.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <string>
#include <exception>
#include <stdexcept>
#include "glog/logging.h"
#include "socket_wrapper.h"

namespace kinetic {

using std::string;

SocketWrapper::SocketWrapper(const std::string& host, int port, bool use_ssl, bool nonblocking)
        : ctx_(nullptr), ssl_(nullptr), host_(host), port_(port), nonblocking_(nonblocking), fd_(-1) {
    if(!use_ssl) return;

    SSL_library_init();
    OpenSSL_add_all_algorithms();
    ctx_ = SSL_CTX_new(SSLv23_client_method());
    ssl_ = SSL_new(ctx_);
    if(!ssl_ || !ctx_)
        throw std::runtime_error("Failed Setting up SSL environment.");
    SSL_set_mode(ssl_, SSL_MODE_AUTO_RETRY);
}

SocketWrapper::~SocketWrapper() {
    if (fd_ == -1) {
        LOG(INFO) << "Not connected so no cleanup needed";
    } else {
        LOG(INFO) << "Closing socket with fd " << fd_;
        if (close(fd_)) {
            PLOG(ERROR) << "Error closing socket fd " << fd_;
        }
    }
    if(ssl_) SSL_free(ssl_);
    if(ctx_) SSL_CTX_free(ctx_);
}

bool SocketWrapper::Connect() {
    LOG(INFO) << "Connecting to " << host_ << ":" << port_;

    struct addrinfo hints;
    memset(&hints, 0, sizeof(struct addrinfo));

    // could be inet or inet6
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_NUMERICSERV;

    struct addrinfo* result;

    string port_str = std::to_string(port_);

    if (int res = getaddrinfo(host_.c_str(), port_str.c_str(), &hints, &result) != 0) {
        LOG(ERROR) << "Could not resolve host " << host_ << " port " << port_ << ": "
                << gai_strerror(res);
        return false;
    }

    struct addrinfo* ai;
    int socket_fd;
    for (ai = result; ai != NULL; ai = ai->ai_next) {
        char host[NI_MAXHOST];
        char service[NI_MAXSERV];
        if (int res = getnameinfo(ai->ai_addr, ai->ai_addrlen, host, sizeof(host), service,
                sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV) != 0) {
            LOG(ERROR) << "Could not get name info: " << gai_strerror(res);
            continue;
        } else {
            LOG(INFO) << "Trying to connect to " << string(host) << " on " << string(service);
        }

        socket_fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);

        if (socket_fd == -1) {
            LOG(WARNING) << "Could not create socket";
            continue;
        }

        // os x won't let us set close-on-exec when creating the socket, so set it separately
        int current_fd_flags = fcntl(socket_fd, F_GETFD);
        if (current_fd_flags == -1) {
            PLOG(ERROR) << "Failed to get socket fd flags";
            close(socket_fd);
            continue;
        }
        if (fcntl(socket_fd, F_SETFD, current_fd_flags | FD_CLOEXEC) == -1) {
            PLOG(ERROR) << "Failed to set socket close-on-exit";
            close(socket_fd);
            continue;
        }

        // On BSD-like systems we can set SO_NOSIGPIPE on the socket to prevent it from sending a
        // PIPE signal and bringing down the whole application if the server closes the socket
        // forcibly
#ifdef SO_NOSIGPIPE
        int set = 1;
        int setsockopt_result = setsockopt(socket_fd, SOL_SOCKET, SO_NOSIGPIPE, &set, sizeof(set));
        // Allow ENOTSOCK because it allows tests to use pipes instead of real sockets
        if (setsockopt_result != 0 && setsockopt_result != ENOTSOCK) {
            PLOG(ERROR) << "Failed to set SO_NOSIGPIPE on socket";
            close(socket_fd);
            continue;
        }
#endif

        if (connect(socket_fd, ai->ai_addr, ai->ai_addrlen) == -1) {
            PLOG(WARNING) << "Unable to connect";
            close(socket_fd);
            continue;
        }

        if (nonblocking_ && fcntl(socket_fd, F_SETFL, O_NONBLOCK) != 0) {
            PLOG(ERROR) << "Failed to set socket nonblocking";
            close(socket_fd);
            continue;
        }

        break;
    }

    freeaddrinfo(result);

    if (ai == NULL) {
        // we went through all addresses without finding one we could bind to
        LOG(ERROR) << "Could not connect to " << host_ << " on port " << port_;
        return false;
    }

    fd_ = socket_fd;
    if(ssl_) return ConnectSSL();
    return true;
}

#include <openssl/err.h>

bool SocketWrapper::ConnectSSL()
{
    SSL_set_fd(ssl_,fd_);
    int rtn = SSL_connect(ssl_);
    if(rtn == 1)
        return true;

    int err = SSL_get_error(ssl_, rtn);
    if( err == SSL_ERROR_WANT_READ || err == SSL_ERROR_WANT_WRITE ){
        fd_set read_fds, write_fds;
        FD_ZERO(&read_fds); FD_ZERO(&write_fds);
        if(err == SSL_ERROR_WANT_READ)  FD_SET(fd_, &read_fds);
        if(err == SSL_ERROR_WANT_WRITE) FD_SET(fd_, &write_fds);
        struct timeval tv = {1,1};
        select(fd_+1, &read_fds, &write_fds, NULL, &tv);
        return ConnectSSL();
    }
    return false;
}

SSL * SocketWrapper::getSSL(){
    return ssl_;
}

int SocketWrapper::fd() {
    return fd_;
}

}  // namespace kinetic
